@@ -7,6 +7,7 @@ use exocore_protos::{
     },
     prost::{Any, ProstAnyPackMessageExt, ProstDateTimeExt},
     store::{entity_query, TraitDetails},
+    test::TestStruct,
 };
 use itertools::Itertools;
 
@@ -20,7 +21,7 @@ use crate::{
 fn fetch_entity_mutations() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let trait1 = IndexOperation::PutTrait(PutTraitMutation {
         block_offset: Some(1),
@@ -102,7 +103,7 @@ fn fetch_entity_mutations() -> anyhow::Result<()> {
 fn search_query_matches() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let now = Utc::now();
     let trait1 = IndexOperation::PutTrait(PutTraitMutation {
@@ -174,7 +175,7 @@ fn search_query_matches() -> anyhow::Result<()> {
 fn search_query_matches_no_fuzzy() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let trait1 = IndexOperation::PutTrait(PutTraitMutation {
         block_offset: Some(1),
@@ -224,7 +225,7 @@ fn search_query_matches_no_fuzzy() -> anyhow::Result<()> {
 fn search_query_matches_recent_boost() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let now = Utc::now();
     let trait1 = IndexOperation::PutTrait(PutTraitMutation {
@@ -308,7 +309,7 @@ fn search_query_matches_recent_boost() -> anyhow::Result<()> {
 fn search_query_matches_paging() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let traits = (0..30).map(|i| {
         let text = "foo ".repeat((i + 1) as usize);
@@ -387,7 +388,7 @@ fn search_query_matches_paging() -> anyhow::Result<()> {
 fn search_query_by_trait_type() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let trait1 = IndexOperation::PutTrait(PutTraitMutation {
         block_offset: None,
@@ -498,7 +499,7 @@ fn search_query_by_trait_type() -> anyhow::Result<()> {
 fn search_query_by_trait_type_paging() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let traits = (0..30).map(|i| {
         IndexOperation::PutTrait(PutTraitMutation {
@@ -563,7 +564,7 @@ fn search_query_by_trait_type_paging() -> anyhow::Result<()> {
 fn sort_by_field() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let traits = (0..20).map(|i| {
         IndexOperation::PutTrait(PutTraitMutation {
@@ -614,7 +615,7 @@ fn sort_by_field() -> anyhow::Result<()> {
 fn search_by_reference() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let et1 = IndexOperation::PutTrait(PutTraitMutation {
         block_offset: None,
@@ -753,7 +754,7 @@ fn search_by_reference() -> anyhow::Result<()> {
 fn search_by_operations() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let et1 = IndexOperation::PutTrait(PutTraitMutation {
         block_offset: None,
@@ -834,7 +835,7 @@ fn search_by_operations() -> anyhow::Result<()> {
 fn search_by_ids() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let et1 = IndexOperation::PutTrait(PutTraitMutation {
         block_offset: None,
@@ -911,7 +912,7 @@ fn search_by_ids() -> anyhow::Result<()> {
 fn search_by_trait_field() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let et1 = IndexOperation::PutTrait(PutTraitMutation {
         block_offset: None,
@@ -966,11 +967,98 @@ fn search_by_trait_field() -> anyhow::Result<()> {
     Ok(())
 }
 
+#[test]
+fn search_by_trait_sub_message_fields() -> anyhow::Result<()> {
+    let registry = Arc::new(Registry::new_with_exocore_types());
+    let config = test_config();
+    let index = MutationIndex::create_in_memory(config, registry)?;
+
+    let et1 = IndexOperation::PutTrait(PutTraitMutation {
+        block_offset: None,
+        operation_id: 1,
+        entity_id: "et1".to_string(),
+        trt: Trait {
+            id: "trt1".to_string(),
+            message: Some(
+                TestMessage {
+                    struct1: Some(TestStruct {
+                        string1: "foo".to_string(),
+                    }),
+                    ..Default::default()
+                }
+                .pack_to_any()?,
+            ),
+            ..Default::default()
+        },
+    });
+    let et2 = IndexOperation::PutTrait(PutTraitMutation {
+        block_offset: None,
+        operation_id: 2,
+        entity_id: "et2".to_string(),
+        trt: Trait {
+            id: "trt2".to_string(),
+            message: Some(
+                TestMessage {
+                    struct1: Some(TestStruct {
+                        string1: "bar".to_string(),
+                    }),
+                    ..Default::default()
+                }
+                .pack_to_any()?,
+            ),
+            ..Default::default()
+        },
+    });
+    index.apply_operations(vec![et1, et2].into_iter())?;
+
+    {
+        // match on any trait should match the sub fields
+        let query = Q::matches("foo");
+        let res = index.search(query.build())?;
+        assert_eq!(res.mutations.len(), 1);
+        find_put_trait(&res, "trt1");
+
+        let query = Q::matches("bar");
+        let res = index.search(query.build())?;
+        assert_eq!(res.mutations.len(), 1);
+        find_put_trait(&res, "trt2");
+    }
+
+    {
+        // can search by field.subfield
+        let query = Q::with_trait_name_query(
+            "exocore.test.TestMessage",
+            TQ::field_equals("struct1.string1", "foo").build(),
+        );
+        let res = index.search(query.build())?;
+        assert_eq!(res.mutations.len(), 1);
+        find_put_trait(&res, "trt1");
+
+        // can search by field prefix
+        let query = Q::with_trait_name_query(
+            "exocore.test.TestMessage",
+            TQ::field_equals("struct1", "foo").build(),
+        );
+        let res = index.search(query.build())?;
+        assert_eq!(res.mutations.len(), 1);
+        find_put_trait(&res, "trt1");
+
+        let query = Q::with_trait_name_query(
+            "exocore.test.TestMessage",
+            TQ::field_equals("struct1", "doesn't exist").build(),
+        );
+        let res = index.search(query.build())?;
+        assert_eq!(res.mutations.len(), 0);
+    }
+
+    Ok(())
+}
+
 #[test]
 fn search_all() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let mut mutations = Vec::new();
     for i in 0..10 {
@@ -1014,7 +1102,7 @@ fn search_all() -> anyhow::Result<()> {
 fn highest_indexed_block() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     assert_eq!(index.highest_indexed_block()?, None);
 
@@ -1079,7 +1167,7 @@ fn highest_indexed_block() -> anyhow::Result<()> {
 fn put_unregistered_trait() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     assert_eq!(index.highest_indexed_block()?, None);
 
@@ -1108,7 +1196,7 @@ fn put_unregistered_trait() -> anyhow::Result<()> {
 fn delete_entity_operation_id_mutation() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let trait1 = IndexOperation::PutTrait(PutTraitMutation {
         block_offset: None,
@@ -1148,7 +1236,7 @@ fn delete_entity_operation_id_mutation() -> anyhow::Result<()> {
 fn put_trait_tombstone() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let contact_mutation = IndexOperation::PutTraitTombstone(PutTraitTombstoneMutation {
         block_offset: None,
@@ -1191,7 +1279,7 @@ fn put_trait_tombstone() -> anyhow::Result<()> {
 fn put_entity_tombstone() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let trait1 = IndexOperation::PutEntityTombstone(PutEntityTombstoneMutation {
         block_offset: None,
@@ -1211,7 +1299,7 @@ fn put_entity_tombstone() -> anyhow::Result<()> {
 fn trait_dates() -> anyhow::Result<()> {
     let registry = Arc::new(Registry::new_with_exocore_types());
     let config = test_config();
-    let mut index = MutationIndex::create_in_memory(config, registry)?;
+    let index = MutationIndex::create_in_memory(config, registry)?;
 
     let creation_date = "2019-08-01T12:00:00Z".parse::<DateTime<Utc>>()?;
     let modification_date = "2019-12-03T12:00:00Z".parse::<DateTime<Utc>>()?;
